#include <stdio.h>
#include "vtty_db.h"
#include "gdbm.h"
#include "errno.h"

static GDBM_FILE dbf;
static int db_count = 0;

int vtty_db_open(int rw)
{
	char dbname[64];

	sprintf(dbname, "%svtty.db", VTTY_DB_PATH);

	do{
		dbf = gdbm_open(dbname, 512, rw, 00664, 0);
	}while(gdbm_errno == GDBM_CANT_BE_READER || gdbm_errno == GDBM_CANT_BE_WRITER);

	if(dbf) return 0;
	else return -1;
}

int vtty_db_ropen(void)
{
	return vtty_db_open(GDBM_READER);
}

int vtty_db_wopen(void)
{
	return vtty_db_open(GDBM_WRCREAT);
}

int vtty_db_wopen_new(void)
{
	return vtty_db_open(GDBM_NEWDB);
}

void vtty_db_close()
{
	if(dbf)
		gdbm_close(dbf);
}

datum struct_to_key(void *p)
{
	datum key = {NULL, 0};
	char keystr[MAXLEN_TTYSUFFIX];

	sprintf(keystr, "%s", ((struct _vtty_db_type_t *)p)->ttysuffix);

	key.dsize = strlen(keystr)+1;
	key.dptr = (void *)malloc(key.dsize);
	if(key.dptr)
		memcpy(key.dptr, keystr, key.dsize);

	return key;
}

datum struct_to_val(void *p)
{
	datum val = {NULL, 0};

	if(p == NULL) return val;

	val.dsize = sizeof(struct _vtty_db_type_t);

	val.dptr = (void *)malloc(val.dsize);
	if(val.dptr) memcpy(val.dptr, p, val.dsize);
	else val.dsize = 0;

	return val;
}
	
int vtty_db_write(void *p)
{
	datum key = {NULL, 0}, val = {NULL, 0};
	int ret = 0;

	key = struct_to_key(p);
	val = struct_to_val(p);

	if(dbf)
	{
		if(gdbm_store(dbf, key, val, GDBM_REPLACE) != 0)
			ret = -1;
		else
			db_count++;
	}
	else
		ret = -1;

	if(key.dptr) free(key.dptr);
	if(val.dptr) free(val.dptr);

	return ret;
}

int vtty_db_get(char *ttysuffix, void *p)
{
	datum key = {ttysuffix, strlen (ttysuffix) + 1};
	datum val;

	if(dbf)
	{
		val = gdbm_fetch(dbf, key);

		if(val.dptr)
		{
			memcpy(p, val.dptr, val.dsize);
			free(val.dptr);
			return 0;
		}
		else
			return -1;
	}
	else
	{
		return -1;
	}
}

int vtty_db_exists(char *ttysuffix)
{
	datum key = {ttysuffix, strlen (ttysuffix) + 1};

	return gdbm_exists(dbf, key);
}

int vtty_db_delete(char *ttysuffix)
{
	int ret = -1;
	datum key = {ttysuffix, strlen(ttysuffix) + 1};
	if(dbf)
	{
		ret = gdbm_delete(dbf, key); // 0 on success, -1 if there is no datum matching with the key
		gdbm_reorganize(dbf);
	}

	if(ret >= 0)
		db_count--;

	return ret;
}


int vtty_db_get_next(void *prevp, void *p)
{
	datum key, prevkey, val;

	if(!dbf)
		return -1;

	if(prevp == NULL)	// get first record
	{
		key = gdbm_firstkey(dbf);
	}
	else
	{
		prevkey = struct_to_key(prevp);
		if(prevkey.dptr == NULL) return -1;
		key = gdbm_nextkey(dbf, prevkey);
		free(prevkey.dptr);
	}

	if(key.dptr)
	{
		val = gdbm_fetch(dbf, key);
		if(val.dptr)
		{
			memcpy(p, val.dptr, val.dsize);
			free(val.dptr);
		}
		free(key.dptr);
		return 0;
	}
	else
	{
		return -1;
	}
}

int vtty_db_count()
{
	return db_count;
}

int vtty_db_get_available_minor()
{
	char inuse[MAX_DB_ENTRY] = {0, };
	struct _vtty_db_type_t data;
	void *p = NULL;
	int i = 0;

	if(vtty_db_count() >= MAX_DB_ENTRY)
		return -1;

	while(vtty_db_get_next(p, (void *)&data) >= 0)
	{
		if(data.minor >= 0 && data.minor < MAX_DB_ENTRY)
			inuse[data.minor] = 1;
		p = &data;
	}

	for(i=0; i<MAX_DB_ENTRY; i++)
	{
		if(inuse[i] == 0) return i;
	}
	return -1;
}

int vtty_db_get_next_by_minornum(const int minor, struct _vtty_db_type_t *data)
{
	void *p = NULL;
	int i = 0;
	int min = MAX_DB_ENTRY;
	struct _vtty_db_type_t mindata;

	if(minor >= MAX_DB_ENTRY)
		return -1;

	while(vtty_db_get_next(p, (void *)data) >= 0)
	{
		if(data->minor == minor && data->minor < MAX_DB_ENTRY)
			return data->minor;
		else if(data->minor > minor && data->minor < min && data->minor < MAX_DB_ENTRY)
		{
			min = data->minor;
			memcpy(&mindata, data, sizeof(struct _vtty_db_type_t));
		}
		p = data;
	}
	if(min > minor && min < MAX_DB_ENTRY)
	{
		memcpy(data, &mindata, sizeof(struct _vtty_db_type_t));
		return min;
	}
	else
		return -1;
}


